"""
**MarkdownX** default URLs, to be added to URLs in the main project.

See URLs in :doc:`../../example` to learn more.
"""
try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url

from .views import (
    ImageUploadView,
    MarkdownifyView,
)


urlpatterns = [
    url('upload/', ImageUploadView.as_view(), name='markdownx_upload'),
    url('markdownify/', MarkdownifyView.as_view(), name='markdownx_markdownify'),
]



import os
import re
from django.test import TestCase
try:
    from django.urls import reverse
except ImportError:  # Djanago < 2.0
    from django.core.urlresolvers import reverse


class SimpleTest(TestCase):

    def test_me(self):
        response = self.client.get('/testview/')
        self.assertEqual(response.status_code, 200)

    def test_upload(self):
        url = reverse('markdownx_upload')
        with open('markdownx/tests/static/django-markdownx-preview.png', 'rb') as fp:
            response = self.client.post(url, {'image': fp}, HTTP_X_REQUESTED_WITH='XMLHttpRequest')

        try:
            json = response.json()
        except AttributeError:  # Django < 1.9
            import json
            json = json.loads(response.content.decode('utf-8'))

        self.assertEqual(response.status_code, 200)
        self.assertIn('image_code', json)

        match = re.findall(r'(markdownx/[\w\-]+\.png)', json['image_code'])
        try:
            if match:
                os.remove(match[0])
        except OSError:
            pass



from django.utils.translation import gettext_lazy as _
from django.forms import ValidationError


class MarkdownxImageUploadError(ValidationError):
    """
    Custom **MarkdownX** exceptions.
    """

    @classmethod
    def not_uploaded(cls):
        """
        No file is available to upload.

        :return: Locale compatible version of the error with the following message:

                 - No files have been uploaded.

        :rtype: MarkdownxImageUploadError
        """
        return cls(_('No files have been uploaded.'))

    @classmethod
    def unsupported_format(cls):
        """

        The file is of a format not defined in :guilabel:`settings.py`
        or if default, in :guilabel:`markdownx/settings.py`.

        :return: Locale compatible version of the error with the following message:

                 - File type is not supported.

        :rtype: MarkdownxImageUploadError
        """
        return cls(_('File type is not supported.'))

    @classmethod
    def invalid_size(cls, current, expected):
        """
        The file is larger in size that the maximum allow in :guilabel:`settings.py` (or the default).

        :param current: Current size of the file.
        :type current: float, int
        :param expected: Expected (maximum permitted) size of the file.
        :type expected: float, int
        :return: Locale compatible version of the error with the following message:

                 - Please keep file size under %(max)s. Current file size: %(current)s.'

        :rtype: MarkdownxImageUploadError
        """
        from django.template.defaultfilters import filesizeformat

        return cls(
            _('Please keep file size under %(max)s. Current file size: %(current)s.') % {
                'max': filesizeformat(expected),
                'current': filesizeformat(current)
            }
        )



from os import path, SEEK_END, SEEK_SET
from io import BytesIO
from uuid import uuid4
from collections import namedtuple

from django import forms
from django.core.files.storage import default_storage
from django.core.files.uploadedfile import InMemoryUploadedFile

from .utils import scale_and_crop, xml_has_javascript
from .exceptions import MarkdownxImageUploadError

from .settings import (
    MARKDOWNX_IMAGE_MAX_SIZE,
    MARKDOWNX_MEDIA_PATH,
    MARKDOWNX_UPLOAD_CONTENT_TYPES,
    MARKDOWNX_UPLOAD_MAX_SIZE,
    MARKDOWNX_SVG_JAVASCRIPT_PROTECTION
)


class ImageForm(forms.Form):
    """
    Used for the handling of images uploaded using the editor through :guilabel:`AJAX`.
    """

    image = forms.FileField()

    # Separately defined as it needs to be
    # processed a text file rather than image.
    _SVG_TYPE = 'image/svg+xml'

    def save(self, commit=True):
        """
        Saves the uploaded image in the designated location.

        If image type is not SVG, a byteIO of image content_type is created and
        subsequently save; otherwise, the SVG is saved in its existing ``charset``
        as an ``image/svg+xml``.

        *Note*: The dimension of image files (excluding SVG) are set using ``PIL``.

        :param commit: If ``True``, the file is saved to the disk;
                       otherwise, it is held in the memory.
        :type commit: bool
        :return: An instance of saved image if ``commit is True``,
                 else ``namedtuple(path, image)``.
        :rtype: bool, namedtuple
        """
        image = self.files.get('image')
        content_type = image.content_type
        file_name = image.name
        image_extension = content_type.split('/')[-1].upper()
        image_size = image.size

        if content_type.lower() != self._SVG_TYPE:
            # Processing the raster graphic image.
            # Note that vector graphics in SVG format
            # do not require additional processing and
            # may be stored as uploaded.
            image = self._process_raster(image, image_extension)
            image_size = image.tell()
            image.seek(0, SEEK_SET)

        # Processed file (or the actual file in the case of SVG) is now
        # saved in the memory as a Django object.
        uploaded_image = InMemoryUploadedFile(
            file=image,
            field_name=None,
            name=file_name,
            content_type=content_type,
            size=image_size,
            charset=None
        )

        if (content_type.lower() == self._SVG_TYPE
                and MARKDOWNX_SVG_JAVASCRIPT_PROTECTION
                and xml_has_javascript(uploaded_image.read())):

            raise MarkdownxImageUploadError(
                'Failed security monitoring: SVG file contains JavaScript.'
            )

        return self._save(uploaded_image, file_name, commit)

    def _save(self, image, file_name, commit):
        """
        Final saving process, called internally after processing tasks are complete.

        :param image: Prepared image
        :type image: django.core.files.uploadedfile.InMemoryUploadedFile
        :param file_name: Name of the file using which the image is to be saved.
        :type file_name: str
        :param commit: If ``True``, the image is saved onto the disk.
        :type commit: bool
        :return: URL of the uploaded image ``commit=True``, otherwise a namedtuple
                 of ``(path, image)`` where ``path`` is the absolute path generated
                 for saving the file, and ``image`` is the prepared image.
        :rtype: str, namedtuple
        """
        # Defining a universally unique name for the file
        # to be saved on the disk.
        unique_file_name = self.get_unique_file_name(file_name)
        full_path = path.join(MARKDOWNX_MEDIA_PATH, unique_file_name)

        if commit:
            default_storage.save(full_path, image)
            return default_storage.url(full_path)

        # If `commit is False`, return the path and in-memory image.
        image_data = namedtuple('image_data', ['path', 'image'])
        return image_data(path=full_path, image=image)

    @staticmethod
    def _process_raster(image, extension):
        """
        Processing of raster graphic image using Python Imaging Library (PIL).

        This is where raster graphics are processed to the specifications
        as defined in ``settings.py``.

        *Note*: The file needs to be uploaded and saved temporarily in the
        memory to enable processing tasks using Python Imaging Library (PIL)
        to take place and subsequently retained until written onto the disk.

        :param image: Non-SVG image as processed by Django.
        :type image: django.forms.BaseForm.file
        :param extension: Image extension (e.g.: png, jpg, gif)
        :type extension: str
        :return: The image object ready to be written into a file.
        :rtype: BytesIO
        """
        thumb_io = BytesIO()
        preped_image = scale_and_crop(image, **MARKDOWNX_IMAGE_MAX_SIZE)
        preped_image.save(thumb_io, extension)
        thumb_io.seek(0, SEEK_END)
        return thumb_io

    @staticmethod
    def get_unique_file_name(file_name):
        """
        Generates a universally unique ID using Python ``UUID`` and attaches the
        extension of file name to it.

        :param file_name: Name of the uploaded file, including the extension.
        :type file_name: str
        :return: Universally unique ID, ending with the extension extracted
                 from ``file_name``.
        :rtype: str
        """
        extension = 1
        extension_dot_index = 1

        file_name = "{unique_name}.{extension}".format(
            unique_name=uuid4(),
            extension=path.splitext(file_name)[extension][extension_dot_index:]
        )
        return file_name

    def clean(self):
        """
        Checks the upload against allowed extensions and maximum size.

        :return: Upload
        """
        upload = self.cleaned_data.get('image')

        # -----------------------------------------------
        # See comments in `self._error_templates` for
        # additional information on each error.
        # -----------------------------------------------
        if not upload:
            raise MarkdownxImageUploadError.not_uploaded()

        content_type = upload.content_type
        file_size = upload.size

        if content_type not in MARKDOWNX_UPLOAD_CONTENT_TYPES:

            raise MarkdownxImageUploadError.unsupported_format()

        elif file_size > MARKDOWNX_UPLOAD_MAX_SIZE:

            raise MarkdownxImageUploadError.invalid_size(
                current=file_size,
                expected=MARKDOWNX_UPLOAD_MAX_SIZE
            )

        return upload



from django.db import models

from .fields import MarkdownxFormField


class MarkdownxField(models.TextField):
    """
    Custom Django field for Markdown text.
    
    Parameters are identical to those of the ordinary Django **TextField** parameters for the model
    fields. For consistency therefore, the descriptions have been borrowed from Django's official
    documentations (with minor modifications).
    
    :param verbose_name: A human-readable name for the field. If the verbose name isn't given,
                         Django will automatically create it using the field's attribute name,
                         converting underscores to spaces.
    :type verbose_name: str
    :param blank: If ``True``, the field is allowed to be blank. Default is ``False``.
    :type blank: bool
    :param null: If ``True``, Django will store empty values as **NULL** in the database.
                 Default is ``False``.
    :type null: bool
    :param editable: If ``False``, the field will not be displayed in the admin or any other
                     ModelForm. They are also skipped during model validation. Default is ``True``.
    :type editable: bool
    :param help_text: Extra "help" text to be displayed with the form widget. It's useful for
                      documentation even if your field isn't used on a form. Note that this value
                      is not HTML-escaped in automatically-generated forms. This lets you include
                      HTML in ``help_text`` if you so desire.
    :type help_text: str
    :param validators: A list of validators to run for this field. See the validators_ documentations
                       for additional information.
    :type validators: list, tuple
    :param error_messages: The error_messages argument lets you override the default messages that the field
                           will raise. Pass in a dictionary with keys matching the error messages you want
                           to override. Relevant error message keys include ``null``, ``blank``,
                           and ``invalid``.
    :type error_messages: dict
    :param kwargs: Other ``django.db.models.field`` parameters. See Django documentations_ for additional
                   information.
    
    .. _documentations: https://docs.djangoproject.com/en/1.10/ref/models/fields/#error-messages
    .. _validators: https://docs.djangoproject.com/en/DEV_CONTENTS_XML/ref/validators/
    """

    def formfield(self, **kwargs):
        """
        Customising the ``form_class``.

        :return: TextField with a custom ``form_class``.
        :rtype: django.db.models.TextField
        """
        defaults = {'form_class': MarkdownxFormField}
        defaults.update(kwargs)
        return super(MarkdownxField, self).formfield(**defaults)



try:
    from django.conf.urls import include, url
except ImportError:
    from django.urls import include, re_path as url

from .views import TestView

urlpatterns = [
    url(r'^testview/', TestView.as_view()),
    url(r'^markdownx/', include('markdownx.urls')),
]
